import sys
from mpl_toolkits.mplot3d import Axes3D
import matplotlib
matplotlib.use("Qt5Agg")
from pathlib import Path
from PyQt5 import QtCore
from PyQt5.QtCore import Qt,QTimer
from PyQt5.QtWidgets import (QApplication, QMainWindow, QMenu, QVBoxLayout, QSizePolicy, QMessageBox, QWidget,QPushButton,QTreeWidget,
                            QTreeWidgetItem, QHBoxLayout, QTextEdit,QAction, QInputDialog,QLineEdit, QFileDialog)
from PyQt5.QtGui import QCursor,QMouseEvent
from CoreWidgets import *
import numpy as np
from yxspkg import tecfile,plot3d
import json
import time
import one2one
import RSCal
#whatsThis explication
'''
blocks:(Blocks)
    block1:(block1)
        surface:I0:(surface:1I0)
Zones:(Zones)
    in:(Zones_child)
        block1:surface:Ke:(zone_surf:surface1Ke)
Rotor and Stator:(RotorStator)
    stator:(Zones_child_stator)
    rotor:(Zones_child_rotor)
        block1:(rs_block)
'''
def ExpandTo3d(key,result,shape):
    #因为face2face返回的结果是针对一个二维的平面，并不包含第三个轴，所以需要转换到和Ogrid_check的结果一样的结果
    idim,jdim,kdim = shape 
    r1,r2,index1,index2 = result
    if key[0] == 'I':
        if key[1] == '0':
            result0 = (1,1), r1, r2, index1+1,index2+1
        else:
            result0 = (idim,idim), r1, r2, index1+1,index2+1
    elif key[0] == 'J':
        if key[1] == '0':
            result0 = r1, (1,1), r2, index1,index2+1
        else:
            result0 = r1, (jdim,jdim), r2, index1,index2+1
    elif key[0] == 'K':
        if key[1] == '0':
            result0 = r1,r2, (1,1),index1,index2
        else:
            result0 = r1,r2, (kdim,kdim),index1,index2
    else:
        raise Exception('key is an error variables x ,y,z not in key.')
    return result0
class YxsQTreeWidgetItem(QTreeWidgetItem):
    def __init__(self,*d,**k):
        super().__init__(*d,**k)
        self.yxsData = dict()
    def GetYxsData(self):
        return self.yxsData
    def AppendYxsData(self,data,key):
        self.yxsData[key] = data
    def GetDataString(self):
        return json.dumps(self.yxsData)
    def setDataFromString(self,dataString):
        self.yxsData = json.loads(dataString)


class BTreeWidget(QTreeWidget):
    def __init__(self,*d):
        super().__init__(*d)
        self.setContextMenuPolicy(Qt.CustomContextMenu)  
        self.customContextMenuRequested.connect(self.show_right_menu)  
  
        self.surface_menu = QMenu(self)  
        self.addtozone_menu = self.surface_menu.addMenu('Added to zone')

        self.zone_surf_menu = QMenu(self)
        action = self.zone_surf_menu.addAction('Delete')
        action.triggered.connect(self.delete_zone_surf)

        self.zones_menu = QMenu(self)
        action = self.zones_menu.addAction('Create')
        action.triggered.connect(self.create_zone)

        self.rotor_menu = QMenu(self)
        action2 = self.rotor_menu.addAction('Create a rotor')
        action2.triggered.connect(self.create_zone_rotor)
        action3 = self.rotor_menu.addAction('Create a stator')
        action3.triggered.connect(self.create_zone_stator)
        self.rotor_menu.addSeparator()
        action4 = self.rotor_menu.addAction('Auto detect boundary')
        action4.triggered.connect(self.AutoDetectBoundary)

        self.zone_menu = QMenu(self)
        action = self.zone_menu.addAction('Delete')
        action.triggered.connect(self.delete_zone)

        self.zone_rotor_menu = QMenu(self)
        action = self.zone_rotor_menu.addAction('Delete')
        action.triggered.connect(self.delete_zone)

        self.block_menu = QMenu(self)
        self.addfacestozone_menu = self.block_menu.addMenu('Added faces to')
        self.addblocktozone_menu = self.block_menu.addMenu('Added block to')


        self.selected_item = None
        self.pressed_state = None
    def mousePressEvent(self,e):
        item = self.itemAt( e.pos())
        if item:
            self.pressed_state = item.checkState(0)
        super().mousePressEvent(e)
    def AutoDetectBoundary(self):
        #自动检测边界
        def One2OneSurf(onedata):
            #将数据转化为[(blockid,'Ke'),,]的形式 blockid从0开始  onedata 的id是从1开始的
            ijk_list = ['I','J','K']
            ijks = [None]*3
            bid,(ijks[0],ijks[1],ijks[2],vdim1,vdim2) = onedata
            vdim = 6 - vdim1 - vdim2
            topBot = '0' if ijks[vdim-1][0] == 1 else 'E'
            return bid - 1, ijk_list[vdim-1]+topBot
        rotor_head = self.father.rotor_head
        zons_set = set([self.father.zones_head.child(i).text(0) for i in range(self.father.zones_head.childCount()) ])
        blocks_head = self.father.blocks_head
        block_find_face = dict(zip(('I0','IE','J0','JE','K0','KE'),range(6)))
        for i in range(rotor_head.childCount()):
            shead = rotor_head.child(i)
            block_ids = [int(shead.child(j).text(0)[5:])-1 for j in range(shead.childCount())]
            if not block_ids:
                continue
            blocks = [self.father.blocks[k] for k in block_ids]
            
            one2oneBlocks = one2one.one2one(blocks)
            #将数据网格块的坐标转化为从1开始的坐标
            one2oneBlocks = [(block_ids[ti-1]+1,tiv,block_ids[tj-1]+1,tjv) for ti,tiv,tj,tjv in one2oneBlocks]
            oneOneSurf = set([One2OneSurf((i[j],i[k])) for i in one2oneBlocks for j,k in [(0,1),(2,3)]])

            


            bound = RSCal.RotorStatorCal(blocks)
            bound = {key:[(block_ids[i],j,k) for i,j,k in v if (block_ids[i],j) not in oneOneSurf] for key,v in bound.items()}
            for mm in bound.items():
                print(mm)

            rotor_name = shead.text(0)
            for bound_key,value in bound.items():
                zname = rotor_name+'-'+bound_key
                if zname not in zons_set:
                    zons_set.add(zname)
                    child = YxsQTreeWidgetItem()
                    child.setText(0,zname)
                    child.setWhatsThis(0,'Zones_child')
                    child.setCheckState(0,Qt.Unchecked)
                    self.father.zones_head.addChild(child)
                for i,j,_ in value:
                    bl = blocks_head.child(i)
                    self.selected_item = bl.child(block_find_face[j])
                    self.addsurf_to_zone(sender_text = zname)
                #把周期面添加到oneoneblock中
                if bound_key == 'periodic':
                    blockId1, faceId1 = value[0][0:2]
                    blockId2, faceId2 = value[1][0:2]
                    face1 = self.father.GetBlockFace(blockId1, faceId1)
                    face2 = self.father.GetBlockFace(blockId2, faceId2)
                    t = one2one.get_rotated_periodic_faces(face1,face2)
                    print('periodic Face',blockId1, faceId1,blockId2, faceId2,t)
                    if t is not None:
                        t1,t2 = [ExpandTo3d(idn,tt,self.father.blocks[blockId1]['X'].shape) for idn,tt in zip((faceId1,faceId2),t)]
                        one2oneBlocks.append( (blockId1+1,t1,blockId2+1,t2) )
            for ii in one2oneBlocks:
                print(ii)
            shead.AppendYxsData(str(one2oneBlocks),'One2One')
        odata = []
        for i in range(rotor_head.childCount()):
            shead = rotor_head.child(i)
            t = shead.GetYxsData()['One2One']
            t = eval(t)
            odata.extend(t)
        m = one2one.transfer_one2one_str(odata)
        print(m)


            
    
    def delete_zone(self):
        self.selected_item.parent().removeChild(self.selected_item)
    def create_zone(self,e,title='Create a zone',label = 'zone name', whatsThis = 'Zones_child',head=None):
        text, assure = QInputDialog.getText(self,title, label, QLineEdit.Normal,"")
        if assure:
            if head is None:
                head = self.father.zones_head
            child = YxsQTreeWidgetItem()
            child.setText(0,text)
            child.setWhatsThis(0,whatsThis)
            child.setCheckState(0,Qt.Unchecked)
            head.addChild(child)
    def create_zone_rotor(self):
        self.create_zone(True,title='Create a rotor', label='rotor name', whatsThis='Zones_child_rotor',head=self.father.rotor_head)
    def create_zone_stator(self):
        self.create_zone(True,title='Create a stator', label='stator name', whatsThis='Zones_child_stator',head=self.father.rotor_head)
    def delete_zone_surf(self):
        #删除zone里面的surface
        self.selected_item.parent().removeChild(self.selected_item)
    def setFather(self,p):
        self.father = p
    def show_right_menu(self,pos):
        father = self.father
        item = self.itemAt(pos)
        if item is None:
            return
        whatsThis = item.whatsThis(0)
        pos = QCursor.pos()
        pos.setX(pos.x()+2)
        pos.setY(pos.y()+2)
        self.selected_item = item
        if whatsThis.startswith('surface'):
            self.addtozone_menu.clear()
            zone_head = father.zones_head
            for i in range(zone_head.childCount()):
                text = zone_head.child(i).text(0)
                action = self.addtozone_menu.addAction(text)
                action.triggered.connect(self.addsurf_to_zone)
            
            self.surface_menu.exec_(pos)
        elif whatsThis.startswith('zone_surf'):
            #右键点击的是zone里面的surface
            self.zone_surf_menu.exec_(pos)
        elif whatsThis == 'Zones':
            #右键点击的是Zones标签
            self.zones_menu.exec_(pos)
        elif whatsThis == 'Zones_child':
            #右键点击的是zone标签
            self.zone_menu.exec_(pos)
        elif whatsThis.startswith('block'):
            #右键点击的是某个block
            self.addfacestozone_menu.clear()
            zone_head = father.zones_head
            for i in range(zone_head.childCount()):
                text = zone_head.child(i).text(0)
                action = self.addfacestozone_menu.addAction(text)
                action.triggered.connect(self.addsurfs_to_zone)
            self.addblocktozone_menu.clear()
            zone_head = father.rotor_head
            for i in range(zone_head.childCount()):
                text = zone_head.child(i).text(0)
                action = self.addblocktozone_menu.addAction(text)
                action.triggered.connect(self.AddBlockToRotor)
            self.block_menu.exec_(pos)
        elif whatsThis == 'RotorStator':
            self.rotor_menu.exec_(pos)
        elif whatsThis == 'Zones_child_rotor' or whatsThis == 'Zones_child_stator':
            self.zone_rotor_menu.exec_(pos)
    def AddBlockToRotor(self):
        text = self.sender().text()
        block_item = self.selected_item
        head = self.father.rotor_head
        for i in range(head.childCount()):
            rotor = head.child(i)
            if rotor.text(0) == text:
                child = YxsQTreeWidgetItem()
                tt = block_item.text(0) 
                child.setText(0,tt)
                child.setWhatsThis(0,'rs_block')
                child.setCheckState(0,QtCore.Qt.Unchecked)
                rotor.addChild(child)
                break
    def addsurf_to_zone(self,*d,**kargs):
        #将blocks中的面添加到zone中，为了以后设置边界条件使用
        zone_head = self.father.zones_head
        surf_item = self.selected_item
        sender_text = kargs.get('sender_text',self.sender().text())
        for i in range(zone_head.childCount()):
            zone = zone_head.child(i)
            text = zone.text(0)
            if text == sender_text:
                zone_surf_whatsThis_set = [zone.child(i).whatsThis(0) for i in range(zone.childCount())]
                this_surf_whatsThis =  'zone_surf:'+surf_item.whatsThis(0)
                if this_surf_whatsThis not in zone_surf_whatsThis_set:
                    child = YxsQTreeWidgetItem()
                    tt = surf_item.parent().text(0) +':' + surf_item.text(0)
                    child.setText(0,tt)
                    child.setWhatsThis(0,this_surf_whatsThis)
                    child.setCheckState(0,QtCore.Qt.Unchecked)
                    zone.addChild(child)
                break
    def addsurfs_to_zone(self):
        block_item = self.selected_item
        for i in range(block_item.childCount()):
            self.selected_item = block_item.child(i)
            self.addsurf_to_zone()
class SettingWidget(QWidget):
    def setFather(self,fa):
        self.father = fa
class ApplicationWindow(QMainWindow):
    def __init__(self):
        QMainWindow.__init__(self)
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.setWindowTitle("程序主窗口")

        self.pwd = Path()  #设置工作路径

        self.file_menu = QMenu('&File', self)
        self.file_menu.addAction('&Open', self.fileOpen,
                                 QtCore.Qt.CTRL + QtCore.Qt.Key_O)
        self.file_menu.addAction('&Save', self.fileSave,
                                 QtCore.Qt.CTRL + QtCore.Qt.Key_S)
        self.file_menu.addSeparator()
        self.file_menu.addAction('&Test', self.test_func,
                                 QtCore.Qt.CTRL + QtCore.Qt.Key_T)
        self.file_menu.addSeparator()
        self.file_menu.addAction('&Quit', self.fileQuit,
                                 QtCore.Qt.CTRL + QtCore.Qt.Key_Q)
        
        self.menuBar().addMenu(self.file_menu)
        self.surface_set_drawed = set()
        self.help_menu = QMenu('&Help', self)
        self.menuBar().addSeparator()
        self.menuBar().addMenu(self.help_menu)

        self.help_menu.addAction('&About', self.about)

        self.main_widget = QWidget(self)

        l = QVBoxLayout()
        window = MplCanvas(self.main_widget, width=10, height=7, dpi=100)
        window.setMinimumSize(100,100)
        l.addWidget(window,8)
        self.window = window
        self.shell = SongziShell()
        self.shell.setMinimumHeight(30)
        l.addWidget(self.shell,1)
        self.blocks_tree = BTreeWidget()
        self.blocks_tree.setFather(self)
        self.blocks_tree.setHeaderLabels(['Grid'])

        self.blocks_tree.setMinimumWidth(200)  # 左边栏
        h = QHBoxLayout(self.main_widget)
        h.addWidget(self.blocks_tree,1)
        self.setting_widget = SettingWidget()
        h.addWidget(self.setting_widget,1)
        self.setting_widget.setMinimumWidth(200)

        h.addLayout(l)

        self.main_widget.setFocus()
        self.setCentralWidget(self.main_widget)
        # 状态条显示2秒
        self.statusBar().showMessage("welcome to cfl3d_pyui", 2000)
        self.delete_display_surface_timer = QTimer(self)
        self.delete_display_surface_timer.timeout.connect(self.delete_display_surf_time_connect)

        self.append_display_surface_timer = QTimer(self)
        self.append_display_surface_timer.timeout.connect(self.append_display_surf_time_connect)
        self.surface_set_append = set()
        self.saved_tree_list = None
        self.child_head_list = list()
        

        self.resize(1200,700)
        self.after_open(self.pwd / 'cfl3d.out')
        
    def test_func(self):
        self.after_open('/home/yxs/F/learning/CFL3D/CFL3D/cfl3d_test_case_nonin/cfl3d.out')
    def append_display_surf_time_connect(self):
        self.append_display_surface_timer.stop()
        self.surface_set_drawed.update(self.surface_set_append)
        for block_id,name,nn in self.surface_set_append:
            block = self.blocks[block_id-1]
            x, y, z = block['X'], block['Y'], block['Z']
            if name == 'I':
                surface = x[nn], y[nn], z[nn]
            elif name == 'J':
                surface = x[:, nn], y[:, nn], z[:, nn]
            elif name == 'K':
                surface = x[:, :, nn], y[:, :, nn], z[:, :, nn]
            self.draw_surface(surface)
        self.surface_set_append.clear()
        self.window.draw()
    def delete_display_surf_time_connect(self):
        self.delete_display_surface_timer.stop()
        for block_id,name,nn in self.surface_set_drawed:
            block = self.blocks[block_id-1]
            x, y, z = block['X'], block['Y'], block['Z']
            if name == 'I':
                surface = x[nn], y[nn], z[nn]
            elif name == 'J':
                surface = x[:, nn], y[:, nn], z[:, nn]
            elif name == 'K':
                surface = x[:, :, nn], y[:, :, nn], z[:, :, nn]
            self.draw_surface(surface)
        
        self.window.draw()
    def SelectedSurfaceOfBlocks(self,item):
        whatsThis = item.whatsThis(0)
        text = whatsThis[8:]
        blockId = int(text[:-2])
        faceId = text[-2:]
        block = self.blocks[blockId-1]
        idim,jdim,kdim = block['X'].shape
        print(idim,jdim,kdim)

    def TreeItemSelected(self,item):
        #树的节点 被单击选中
        whatsThis = item.whatsThis(0)
        if whatsThis.startswith('surface'):
            self.SelectedSurfaceOfBlocks(item)
    def blocks_change(self,item,column):
        def change_state(item,state,column):
            item.setCheckState(0,state)
            self.blocks_change(item,column)
        whatsThis = item.whatsThis(0)
        state = item.checkState(0)
        if item.isSelected():
            #如果是单击选中而不是勾选 则执行以下内容
            self.TreeItemSelected(item)
            if self.blocks_tree.pressed_state == state:
                return
        if whatsThis == 'Blocks' or whatsThis == 'Zones' or whatsThis.startswith('block') or whatsThis.startswith('Zones_child'):
            [change_state(item.child(i),state,column) for i in range(item.childCount())]
        elif whatsThis.startswith('surface') or whatsThis.startswith('zone_surf:'):
            if whatsThis.startswith('zone_surf:'):
                whatsThis = whatsThis[10:]
            block_id = int(whatsThis[8:-2])
            surface_name = whatsThis[-2:]
            if state == Qt.Checked:
                self.plot_add_surface(block_id,surface_name)
            else:
                self.plot_delete_surface(block_id, surface_name)

    def plot_add_surface(self,block_id,surface_name):
        """block_id:int'
           surface_name:str,'surface:I0'
        """ 
        
        if surface_name[-1] == '0':
            nn = 0
        elif surface_name[-1].upper() == 'E':
            nn = -1
        else:
            raise Exception('surface_Name Error',surface_name)
        s = block_id,surface_name[-2],nn
        if s not in self.surface_set_append and s not in self.surface_set_drawed:
            self.surface_set_append.add(s)
            if not self.append_display_surface_timer.isActive():
                self.append_display_surface_timer.start(50)
    def plot_delete_surface(self,block_id,surface_name):
        if surface_name[-1] == '0':
            nn = 0
        elif surface_name[-1].upper() == 'E':
            nn = -1
        s = block_id,surface_name[-2],nn
        self.surface_set_drawed.discard(s)
        self.ax.clear()
        if not self.delete_display_surface_timer.isActive():
            self.delete_display_surface_timer.start(50)
        
    def GetBlockFace(self,blockid,faceid):
        #blockid: int start from 0 
        #faceid : Ke
        p = 0 if faceid[1] == '0' else -1
        block = self.blocks[blockid]
        x,y,z = block['X'], block['Y'], block['Z']
        if faceid[0] == 'I':
            face = x[p],    y[p],    z[p]
        elif faceid[0] == 'J':
            face = x[:,p],  y[:,p],  z[:,p]
        elif faceid[0] == 'K':
            face = x[:,:,p],y[:,:,p],z[:,:,p]
        return face
    def set_blocks_tree(self):
        
        self.blocks_tree.itemClicked.connect(self.blocks_change)

        self.blocks_tree.clear()

        self.blocks_head = YxsQTreeWidgetItem(self.blocks_tree)
        self.blocks_head.setText(0,'Blocks')
        self.blocks_head.setWhatsThis(0,'Blocks')
        self.blocks_head.setCheckState(0,QtCore.Qt.Unchecked)

        for i,_ in enumerate(self.blocks):
            child = YxsQTreeWidgetItem()
            child.setText(0,'block'+str(i+1))
            child.setWhatsThis(0,'block'+str(i+1))
            child.setCheckState(0,QtCore.Qt.Unchecked)
            self.blocks_head.addChild(child)
            for j in ['I0','IE','J0','JE','K0','KE']:
                surf_child = YxsQTreeWidgetItem()
                surf_child.setText(0,'surface:{}'.format(j))
                surf_child.setWhatsThis(0,'surface:'+str(i+1)+j)
                surf_child.setCheckState(0,QtCore.Qt.Unchecked)
                child.addChild(surf_child)
        self.zones_head = YxsQTreeWidgetItem(self.blocks_tree)
        self.zones_head.setText(0,'Zones')
        self.zones_head.setWhatsThis(0,'Zones')
        self.zones_head.setCheckState(0,QtCore.Qt.Unchecked)

        self.rotor_head = YxsQTreeWidgetItem(self.blocks_tree)
        self.rotor_head.setText(0,'Rotor and Stator')
        self.rotor_head.setWhatsThis(0,'RotorStator')
        self.rotor_head.setCheckState(0,QtCore.Qt.Unchecked)

        self.child_head_list = [self.zones_head,self.rotor_head]

        if not self.saved_tree_list:
            for i in ['in','out']:
                child = YxsQTreeWidgetItem()
                child.setText(0,i)
                child.setWhatsThis(0,'Zones_child')
                child.setCheckState(0,QtCore.Qt.Unchecked)
                self.zones_head.addChild(child)
        else:
            for head,tree_dict in zip(self.child_head_list, self.saved_tree_list):
                self.Dict2Tree(head,tree_dict)
    
    def start_draw_plot3d_grid(self):
        self.window.fig.clear()
        axes = self.window.fig.gca(projection='3d')
        if self.plot3d_file.suffix == '.xs3d':
            fp = open(self.plot3d_file,'rb')
            tt = plot3d.read_plot3d_unfmt(fp)
            dict_str = fp.read().decode('utf8')
            self.saved_tree_list = json.loads(dict_str)
        else:
            self.saved_tree_list = None
            tt = tecfile.read(str(self.plot3d_file))

        self.blocks = tt
        self.ax = axes 
        self.is_draw_surfaces = True
        self.set_blocks_tree()
        self.window.fig.subplots_adjust(bottom = 0,left=0,top=1,right=1)
        # self.ax.get_proj = lambda: np.dot(Axes3D.get_proj(self.ax), np.diag([0.5, 1, 1, 1]))
        # x,y,z = tt[0]['X'][0],tt[0]['Y'][0],tt[0]['Z'][0]
        # self.draw_surface((x,y,z))
    def draw_surface(self,surface):
        if not self.is_draw_surfaces:
            return 
        points_n = 15
        x,y,z = surface
        h,w = x.shape
        inx = np.array(range(0, h, min(h, int(h/points_n)+1)),dtype='int32')
        iny = np.array(range(0, w, min(w, int(w/points_n)+1)), dtype='int32')
        inx[-1] = h-1
        iny[-1] = w-1
        x1 = x[inx][:, iny]
        y1 = y[inx][:, iny]
        z1 = z[inx][:, iny]
        self.ax.set_xlabel('X')
        self.ax.set_ylabel('Y')
        self.ax.set_zlabel('Z')
        
        rgb = np.where(x,'#BFAB6EC0','#BFAB6EC0')
        self.ax.plot_surface(x1, y1, z1, rstride=1, cstride=1, facecolors=rgb,
                            linewidth=0, antialiased=True)
        self.ax.plot(x[0],y[0],z[0],linewidth = 3, color = 'k')
        self.ax.plot(x[-1],y[-1],z[-1],linewidth = 3, color = 'k')
        self.ax.plot(x[:,0],y[:,0],z[:,0],linewidth = 3, color = 'k')
        self.ax.plot(x[:,-1],y[:,-1],z[:,-1],linewidth = 3, color = 'k')
        # print(len(self.ax.lines))
    def fileOpen(self):
        fileName,_ = QFileDialog.getOpenFileName(self, "Open File")
        if fileName:
            self.after_open(fileName)
    def fileSave(self):
        fileName,ext_f= QFileDialog.getSaveFileName(self, "Save File","","plot3d ascii (*.fmt);;plot3d bin (*.g);;xs3d bin (*.xs3d)")
        if fileName:
            m,n = ext_f.find('*'), ext_f.find(')')
            ext_f = ext_f[m+1:n]
            if not fileName.endswith(ext_f):
                fileName += ext_f
            if ext_f == '.xs3d':
                self.save_xs3d(fileName)
            else:
                tecfile.write(fileName,self.blocks)
    def save_xs3d(self,fileName):
        fp = open(fileName,'wb')
        plot3d.write_plot3d_unfmt(fp,self.blocks)
        tree_dict_list = [self.Tree2Dict(head) for head in self.child_head_list]
        dict_str = json.dumps(tree_dict_list)
        fp.write(dict_str.encode('utf8'))
        fp.close()
    def Dict2Tree(self,head,children):
        for key,value in children.items():
            child = YxsQTreeWidgetItem()
            child.setText(0,key)
            child.setWhatsThis(0,value['whatsThis'])
            child.setDataFromString(value['data'])
            child.setCheckState(0,QtCore.Qt.Unchecked)
            head.addChild(child)
            if value['children']:
                self.Dict2Tree(child,value['children'])
    def Tree2Dict(self,tree):
        d = dict()
        for i in range(tree.childCount()):
            child = tree.child(i)
            text, whatsThis = child.text(0), child.whatsThis(0)
            child_dict = dict()
            child_dict['whatsThis'] = whatsThis 
            child_dict['data'] = child.GetDataString()
            child_dict['children'] = self.Tree2Dict(child)
            d[text] = child_dict 
        return d

    def after_open(self,filename):
        pname = Path(filename)
        if pname.suffix == '.out':
            self.residual_file = pname
            self.start_draw_residual()
        else:
            self.plot3d_file = pname 
            self.start_draw_plot3d_grid()
    def fileQuit(self):
        self.close()

    def closeEvent(self, ce):
        self.fileQuit()

    def about(self):
        QMessageBox.about(self, "About",
        """
        Copyright 2018 Yongxiaosong 
        """
        )
    def start_draw_residual(self):
        # return
        self.window.fig.clear()
        self.last_data_draw = dict()
        self.draw_history = True #是否绘制历史残差
        if not self.residual_file.is_file():
            return
        self.residual = Residual(self.residual_file)
        self.res_timer = QtCore.QTimer(self)
        self.res_timer.timeout.connect(self.draw_residual)
        self.res_timer.start(2000)
        self.axes = self.window.fig.add_subplot(111)
        self.window.fig.subplots_adjust(bottom = 0.05,left=0.1,top=1,right=1)
        self.need_legend = True
    def draw_residual(self):
        if self.draw_history:
            data = self.residual.get_history()
        else:
            data = np.array([])
        if not data.any():
            data = self.residual.get()
        if not data.any():return
        self.draw_history = False
        x = data[:,2]
        fp = 3
        x_last, y_last = None, None
        colors = ['k','b','y','#054E9F','g','o']
        names = ('residual','total res.')
        for i,(key,co) in enumerate(zip(names,colors)):
            #为了使图线不间断
            if key not in self.last_data_draw:
                x0 = x 
                y0 = np.abs(data[:,i+fp])
                self.last_data_draw[key] = x0[-1], y0[-1]
            else:
                x_last, y_last = self.last_data_draw[key]
                x0 = np.concatenate(([x_last],x))
                y0 = np.concatenate(([y_last],np.abs(data[:,i+fp])))
                self.last_data_draw[key] = x0[-1], y0[-1]

            self.axes.semilogy(x0,y0,label = key,color = co)

        if self.need_legend:
            self.axes.legend(loc='upper left')
            self.need_legend=False
        self.window.draw()



if __name__ == '__main__':
    app = QApplication(sys.argv)
    aw = ApplicationWindow()
    aw.setWindowTitle("CFL3D Tools")
    aw.show()
    sys.exit(app.exec_())
    # app.exec_()



import pexpect
import sys 
import os
def run(terminal = True):
    def deal(command):
        command = command.strip()
        n = command.find(':')
        if n !=-1:
            command = command[:n+1]+'/home/public/yxs/'+command[n+1:]
            command = 'public@'+command 
            return command
        else:
            return command
    v = [deal(i) for i in sys.argv[1:]]
    com ='scp '+ ' '.join(v)
    print(com)
    if terminal:
        os.system(com)
    else:
        t = pexpect.spawn(com)
        f = t.expect('.password:')
        if f==0:
            t.sendline('public')
            t.read()
def tianhe():
    for i in sys.argv[1:]:
        if i.find('blsc605:')!=-1:
            break 
    else:
        return False
    command = ' '.join(sys.argv[1:]).replace('blsc605:','blsc605@gz:/PARA/blsc605/')
    command = '/home/yxs/F/learning/tianhe/papp_cloud-2.6.11/papp_cloud_linux64 scp -i /home/yxs/F/learning/tianhe/papp_cloud-2.6.11/blsc605.id '+command 
    print(command)
    os.system(command)
    return True

if __name__=='__main__':
    is_tianhe = tianhe()
    if not is_tianhe:
        run()


from pathlib import Path
import time
import numpy as np 
from functools import wraps
import re
from matplotlib import pyplot as plt
import sys
def result_wrapper(func,last_info=[0,1]):
    @wraps(func)
    def wrapper(*args, **kwargs):
        t = func(*args, **kwargs)
        if t is None:
            return np.array([])
        t = [i.split() for i in t]
        if t and not t[-1][2].isdigit():
            if t[0][2].isdigit():
                kt = t[-1][2]
                last_digit = 0
                for i,v in enumerate(t):
                    if v[2] == kt:
                        last_digit = i - 1
                        break 
                delta = int(t[last_digit][2]) - int(t[last_digit-1][2])
            else:
                last_digit,delta = last_info
            for i in t[last_digit+1:]:
                last_digit += delta
                i[2] = last_digit

            last_info[0] = last_digit
            last_info[1] = delta

        try:
            t = np.array(t,dtype='float32')
        except:
            print(t)
        if not t.any():
            return t
        self = args[0]
        if self.blk is None:
            #取self.blk 为blk序列中的众数
            m = t[:,1]
            m = m[1:] - m[:-1]
            m = t[:,1][np.where(m==0)]
            self.blk = m[0]

        t = t[np.where(np.abs(t[:,1] - self.blk)<0.01)]
        return t
    return wrapper
class Residual:
    def __init__(self,filename):
        self.filename = Path(filename)
        self.pos=-1
        self.titles = None
        self.re_res = None
        self.latest_mtime = 0
        self.re_res = self.get_re_res()
        self.blk = None
        self.iter_last = 0
    def get_re_res(self):
        fp = open(self.filename,'r')
        for i in fp:
            if i.find('total res.')!=-1:
                self.titles = re.split(' {2,}',i)
                n = len(self.titles) - 3
                return re.compile('\s+\d+\s+\d+\s+(\d+|\*{{6}})\s+([-\.\d]+E[\+-]\d+\s+){{{n}}}'.format(n = n))
        return None

    @result_wrapper
    def get(self):
        mtime = self.filename.lstat().st_mtime
        if mtime <= self.latest_mtime:
            return []
        self.latest_mtime = mtime
        fp = open(self.filename,'r')
        if self.re_res is None:
            t = self.get_re_res()
            if t is not None:
                self.re_res = t
            else:
                return None
        fp.seek(self.pos+1,0)
        res_list = [i for i in fp if self.re_res.match(i)]
        self.pos = fp.tell()
        return res_list
    def get_history(self):
        fp = open(self.filename,'r')
        is_title = False
        title_re = re.compile(' *iteration *residual *CL *CD *CY *CMY.*')
        data_re  = re.compile('( +\d+)( +[\.\dEe\+\-]+)( +[\.\dEe\+\-]+)( +[\.\dEe\+\-]+)( +[\.\dEe\+\-]+)( +[\.\dEe\+\-]+)')
        result = []
        for line in fp:
            if title_re.match(line):
                is_title = True 
                continue
            if is_title==True:
                t = data_re.findall(line)
                if t:
                    result.append(t)
                else:
                    break
        if not is_title:
            return np.array([])
        result = np.array(result,dtype = 'float32').reshape((len(result),-1))
     

        result = np.hstack((result[:,-2:],result,(result[:,-2:])))
        result[:,4] = result[:,3]
        return result

def run(filename,interval):
    plt.ion()
    plt.figure(filename)
    res = Residual(filename)
    k = res.get_history()
    is_first = True
    min_shape = None
    try:
        while True:
            k2 = res.get()
            if is_first:
                if k.any():
                    min_shape = min(k.shape[1],k2.shape[1])
                    k = np.vstack((k[:,:min_shape],k2[:,:min_shape]))
                else:
                    k = k2
                    min_shape = k2.shape[1]
            
            else:
                if not k2.any():
                    plt.pause(2*interval)
                    continue
                else:
                    k = np.vstack((k[-1:,:min_shape],k2[:,:min_shape]))
            x = k[:,2]
            y1 = k[:,3]
            y2 = k[:,4]
            plt.semilogy(x,y1,'k-',label='$residual$')
            plt.semilogy(x,y2,'b-',label = '$total res.$')
            l = k[-1]
            titles = res.titles
            for i in range(5,k.shape[1]):
                if not titles:break
                if titles[i] == 'time':break
                if abs(l[i]) > 1e-18:
                    y = np.abs(k[:,i])
                    plt.semilogy(x,y,label='${}$'.format(titles[i]))
            plt.xlabel('$iterations$')
            plt.ylabel('$res$')
            if is_first:
                plt.legend()
            plt.pause(interval)
            is_first = False
    except:
        pass
def run_sub(filename,interval):
    plt.figure(filename)
    fp = open(filename)
    fp.readline()
    xy = np.array([i.split()[:2] for i in fp],dtype='float32')
    plt.plot(xy[:,0],xy[:,1])
    plt.show()
def find_parm(p):
    for i,v in enumerate(sys.argv):
        if v.strip() == p:
            if p[:2]=='--':return True
            return sys.argv[i+1]
    return None
if __name__ == '__main__':
    fname = None
    interval = find_parm('-t')
    if interval:
        interval = float(interval)
    else:
        interval = 10
    fname = find_parm('-i')
    if not fname:
        fname = [i.name for i in Path('./').glob('*.out') if i.name != 'precfl3d.out']
        if len(fname)>=2:
            print('Warning: too many out files',fname)
        if fname:
            fname = fname[0]
    fsub = find_parm('--s') 
    if fsub:
        fsub = 'cfl3d.subit_res'
        run_sub(fsub,interval)
    else:
        run(fname,interval)


    



from pathlib import Path
import json
from . import data_format as df
from . import interpolate
from yxspkg import tecfile
import sys
import numpy as np
from . import RSCal
def get_info(inpfile):
    print(inpfile)
    fp = open(inpfile)
    fp.readline()
    mesh_file = fp.readline().strip()
    output_mesh = fp.readline().strip()
    output_q = fp.readline().strip()
    info = {'mesh_file':mesh_file,'output_mesh':output_mesh,"output_q":output_q}
    while True:
        s = fp.readline().lower()
        if s.find("mach")!=-1 and s.find('ialph')!=-1:
            data = [float(i) for i in fp.readline().split()]
            info['mach'] = data[0]
            info['reue'] = data[3]
            info['T'] = data[4]/9*5
            break 
    json_data = ''
    for i in fp:
        if i.startswith('{'):
            json_data = i
            break
    if json_data:
        json_data += fp.read()
        json_data = json.loads(json_data)
        info.update(json_data)
    return info
    
def cal_area_zone(zone,xtag = 'CoordinateX', ytag = 'CoordinateY', ztag='CoordinateZ'):
    def cal_area(p1,p2,p3,p4):
        v1 = np.cross(p2-p1,p3-p1)*0.5
        v1+= np.cross(p3-p1,p4-p1)*0.5 
        s  = ((v1**2).sum(axis=-1)**0.5)[:,:,:,None]
        return np.concatenate((v1,s),axis=-1)
    t=np.concatenate((zone[xtag][:,:,:,None],zone[ytag][:,:,:,None],zone[ztag][:,:,:,None]),axis=-1)
    p1 = t[:-1,:-1,:-1]
    p2 = t[ 1:,:-1,:-1]
    p3 = t[ 1:, 1:,:-1]
    p4 = t[:-1, 1:,:-1]
    p5 = t[:-1,:-1, 1:]
    p6 = t[ 1:,:-1, 1:]
    p7 = t[ 1:, 1:, 1:]
    p8 = t[:-1, 1:, 1:]
    s1 = cal_area(p1,p4,p8,p5)
    s2 = cal_area(p2,p3,p7,p6)
    s3 = cal_area(p2,p1,p5,p6)
    s4 = cal_area(p3,p4,p8,p7)
    s5 = cal_area(p1,p2,p3,p4)
    s6 = cal_area(p5,p6,p7,p8)
    return s1,s2,s3,s4,s5,s6
def main():
    if len(sys.argv)>1:
        inpfile = Path(sys.argv[1] )
    else:
        inps = [i for i in Path('./').glob('*.inp') if i.name != 'tempz3y2x1.inp']
        if len(inps) > 1:
            print("which file?")
            for i,f in enumerate(inps):
                print(f"{i+1}: {f}")
            x = input("type the number:")
            inpfile = inps[int(x)-1]
        elif len(inps)==0:
            print("no available file")
            inpfile = None
        else:
            inpfile = inps[0]
    if not inpfile:
        return
    info = get_info(inpfile)
    togrid = info.get('togrid',None)
    if togrid:
        patterns = interpolate.get_patterns(str(inpfile),vectors=info.get('vectors',[]))
        is_double = info.get("double",False)
    while True:
        datafile = info['output_q']
        outfile = str(info.get('output','*.szplt'))
        if outfile.find('*')!=-1:
            stem = Path(datafile).stem 
            outfile = outfile.replace('*',stem)
        qdata = tecfile.read(datafile)
        if togrid:
            qdata = interpolate.interpolate_center2grid(qdata,patterns,is_double=is_double)

        if outfile.endswith('szplt'):
            rho = info.get('density',None)
            R = info.get('R',287.14)
            gamma = info.get('gamma',1.4)
            T = info['T']
            if rho:
                P = rho*R*T 
            else:
                P = info['pressure']
            t = df.q2szplt(None,info['mesh_file'],qdata,P,T,gamma=gamma,R=R)
        else:
            t = qdata
        tecfile.write(outfile,t)
        break
    return outfile
def center_value(data):
    return 0.25*(data[1:, 1:] + data[:-1, 1:] + data[ :-1, :-1] + data[ 1:, :-1])
def get_area(face,areas):
    blockid,faceSE = face 
    area = areas[blockid]
    faceSE = faceSE.upper()
    if faceSE == 'I0':
        farea = area[0][0,:,:]
    elif faceSE == 'K0':
        farea = area[4][:,:,0]
    elif faceSE == 'KE':
        farea = area[5][:,:,-1]
    else:
        print("error")
        exit() 
    return farea 
def get_var(face,zones,var,farea):
    blockid,faceSE = face 
    zone = zones[blockid]
    value = zone[var]
    faceSE = faceSE.upper()
    if faceSE == 'I0':
        fvar  = value[0,:,:]
    elif faceSE == 'K0':
        fvar  = value[:,:,0]
    elif faceSE == 'KE':
        fvar  = value[:,:,-1]
    else:
        print("error")
        exit() 
    if farea.shape[0] != fvar.shape[0]:
        fvar = center_value(fvar)
    return fvar
def get_mass(face,zones,areas,vrho="Density",vu='VelocityX',vv='VelocityY',vw='VelocityZ'):
    farea = get_area(face,areas)
    rho   = get_var(face,zones,vrho,farea)
    u     = get_var(face,zones,vu  ,farea)
    v     = get_var(face,zones,vv  ,farea)
    w     = get_var(face,zones,vw  ,farea)
    mass  = (u*farea[:,:,0]+v*farea[:,:,1]+w*farea[:,:,2])*rho 
    if face[1][1] == '0':
        mass *= -1
    s = farea[:,:,3].sum()
    return mass,s
def face_sum(face,zones,var,areas,weight='area'):
    farea= get_area(face,areas)
    fvar = get_var(face,zones,var,farea)
    if weight == 'area':
        farea = farea[:,:,3]
    elif weight == 'mass':
        farea = get_mass(face,zones,areas)[0]
    return (farea*fvar).sum(),farea.sum()
def multi_faces_average(faces,zones,areas,var='Pressure',weight='area'):
    v0,s0 = 0,0
    for i in faces:
        v,s = face_sum(i[:2],zones,var,areas,weight)
        v0 += v 
        s0 += s 
    return v0/s0
def convert2blocks(zone,xtag = 'CoordinateX', ytag = 'CoordinateY', ztag='CoordinateZ'):
    return {'X':zone[xtag],'Y':zone[ytag],'Z':zone[ztag]}
def cal_mass(faces,zones,areas):
    v0, s0 = 0,0 
    for i in faces:
        v,s = get_mass(i[:2],zones,areas)
        v0 += v.sum()
        s0 += s 
    return v0
def cal_average(zones,stages = None):
    areas = [cal_area_zone(zones[i]) for i in range(len(zones))]
    blocks= [convert2blocks(zones[i]) for i in range(len(zones))]
    t = RSCal.RotorStatorCal(blocks)
    faces_in = t['in']
    faces_out= t['out']
    mass_in = cal_mass(faces_in,zones,areas)
    mass_out= cal_mass(faces_out,zones,areas)
    pt_in = multi_faces_average(faces_in,zones,areas,weight='mass',var='PressureStagnation')
    pt_out = multi_faces_average(faces_out,zones,areas,weight='mass',var='PressureStagnation')
    tt_in = multi_faces_average(faces_in,zones,areas,weight='mass',var='TemperatureStagnation')
    tt_out = multi_faces_average(faces_out,zones,areas,weight='mass',var='TemperatureStagnation')
    k = 1.4 
    pi = pt_out/pt_in 
    coeff = (pi**((k-1)/k)-1)/(tt_out/tt_in-1)
    print('进口总压',pt_in)
    print('出口总压',pt_out)
    print('压比',pi)
    print('进口总温',tt_in)
    print('出口总温',tt_out)
    print('温比',tt_out/tt_in)
    print('效率',coeff)
    print('进口流量', abs(mass_in))
    print('出口流量',abs(mass_out))
    return pt_in,pt_out,pi,coeff,abs(mass_in),abs(mass_out)
if __name__=='__main__':
    outfile = main()
    # cfl3d_case = '/home/yxs/F/learning/CFL3D/323_0409/ok/series/323_new_bound_0001/rotate323_0409_lref1_output.szplt'
    f = tecfile.read(outfile)
    cal_average(f)
    # print(f[0]['CoordinateX'][:,:,:,None])
    


from yxspkg import tecfile 
import numpy as np
import math
import json
import sys
import time
from pathlib import Path
def get_one2one_data(inpfile):
    fp = open(inpfile)
    def getline(line):
        line = line.split()
        grid = int(line[1])
        istart_end = int(line[2]),int(line[5])
        jstart_end = int(line[3]),int(line[6])
        kstart_end = int(line[4]),int(line[7])
        v1 = int(line[8])
        v2 = int(line[9])
        return grid,(istart_end,jstart_end,kstart_end,v1,v2)
    one2oneData = None 
    shapes=[]
    periodicData = []
    file_suffix = inpfile.split('.')[-1]
    if file_suffix =='inp':
        while True:
            line = fp.readline().lower()
            if line.find('idim')!=-1 and line.find('jdim')!=-1 and line.find('kdim') !=-1:
                while True:
                    line = fp.readline().split()
                    if len(line) == 3:
                        shapes.append(tuple([int(i) for i in line]))
                    else:
                        break 
                continue
            if line.find('bctype') != -1 and line.find('ndata') != -1:
                nface = 0
                while True:
                    line = fp.readline().lower()
                    if line.find('mseq') != -1:
                        break 
                    if line.find('bctype') != -1 and line.find('ndata') != -1:
                        nface += 1
                    k = line.split()
                    if len(k) == 8 and k[0].isdigit():
                        if k[2].isdigit() and int(k[2]) == 2005:
                            fp.readline()
                            t = fp.readline().split()
                            periodicData.append((
                                int(k[0]),
                                nface,
                                int(k[3]),
                                int(k[4]),
                                int(k[5]),
                                int(k[6]),
                                int(t[0]),
                                float(t[1]),
                                float(t[2]),
                                float(t[3]),
                            ))

            if line.find('1-1') != -1:
                fp.readline()
                one_length = int(fp.readline())
                fp.readline()
                start = [getline(fp.readline()) for i in range(one_length)]
                fp.readline()
                end = [getline(fp.readline()) for i in range(one_length)]
                one2oneData = [(i1,i2,j1,j2) for (i1,i2),(j1,j2) in zip(start,end)]
                break
    elif file_suffix == 'json':
        json_data = json.load(open(inpfile))
        print(json_data)
    return one2oneData,periodicData,shapes

def get_face_pattern(one2oneData):
    def number_to_slice(i0,i1):
        if i1==i0:
            return i0-1
        elif i1>i0:
            return slice(i0,i1-1)
        else:
            return slice(i0-2,i1-1,-1)
    def parse11(odata,tdata):
        blockid,d1 = odata
        tid,d2 = tdata
        r1 = [number_to_slice(*i) for i in d1[:3]]
        r2 = [number_to_slice(*i) for i in d2[:3]]
        if (d1[-2]-d1[-1])*(d2[-2]-d2[-1]) < 0:
            is_trans = True 
        else:
            is_trans = False 
        return blockid-1,r1,tid-1,r2,is_trans
    return [parse11(i[:2],i[2:]) for i in one2oneData]


def interpolate_center2grid(qdata, patterns ,scalars=[],is_double=False):
    #qdata 以及 qdata【0】都是list数据类型
    if isinstance(qdata[0], dict):
        keys = [i for i in qdata[0] if isinstance(i, int)]
        keys.sort()
        aux_data = [{key:val for key,val in q.items() if isinstance(key, str)} for q in qdata]
        qdata = [[i[k] for k in keys] for i in qdata]
    else:
        keys = None
    patterns = [i if i else [] for i in patterns]
    def interpolate_body(data):
        #插值网格块内的数据
        if is_double:
            fdtype='float64'
        else:
            fdtype='float32'
        temp = np.zeros([i+1 for i in data.shape],dtype=fdtype)
        temp[1:-1, 1:-1, 1:-1] = 0.125*(data[1:, 1:, 1:] + data[1:, :-1, 1:] + data[1:, :-1, :-1] + data[1:, 1:, :-1] +
                                        data[:-1, 1:, 1:] + data[:-1, :-1, 1:] + data[:-1, :-1, :-1] + data[:-1, 1:, :-1])
        #面网格采取4分之一的
        p=(0,-1)
        temp[p,1:-1,1:-1] = (data[p,1:,1:] + data[p,:-1,:-1] + data[p,1:,:-1] + data[p,:-1,1:])/4
        temp[1:-1,p,1:-1] = (data[1:,p,1:] + data[:-1,p,:-1] + data[1:,p,:-1] + data[:-1,p,1:])/4
        temp[1:-1,1:-1,p] = (data[1:,1:,p] + data[:-1,:-1,p] + data[1:,:-1,p] + data[:-1,1:,p])/4
        #棱网格取二分之一，
        i,j = list(zip(*[(i,j) for i in [0,-1] for j in [0,-1]]))
        temp[i,j,1:-1] = (data[i,j,1:] + data[i,j,:-1])/2
        temp[i,1:-1,j] = (data[i,1:,j] + data[i,:-1,j])/2
        temp[1:-1,i,j] = (data[1:,i,j] + data[:-1,i,j])/2
        # 顶点取最近一个点
        i,j,k = list(zip(*[(i,j,k) for i in [0,-1] for j in [0,-1] for k in [0,-1]]))
        temp[i,j,k] = data[i,j,k]
        return temp
    grid_data = [[interpolate_body(data) for data in qs] for qs in qdata]
    face_pattern, line_pattern, point_pattern, periodic_pattern = patterns
    #面网格插值
    for bid1, r1, bid2, r2, is_trans in face_pattern:
        i1,j1,k1 = r1 
        i2,j2,k2 = r2
        for nd in range(len(qdata[0])):
            data1 = grid_data[bid1][nd]
            data2 = grid_data[bid2][nd]
            if is_trans:
                data1[i1,j1,k1] = (data1[i1,j1,k1] + data2[i2,j2,k2].T )/2
                data2[i2,j2,k2] = data1[i1,j1,k1].T
            else:
                data1[i1,j1,k1] =( data1[i1,j1,k1] + data2[i2,j2,k2])/2
                data2[i2,j2,k2] = data1[i1,j1,k1]
    #棱网格插值 和 角点网格插值
    for lines in line_pattern+point_pattern:
        for nd in range(len(qdata[0])):
            t0 = None 
            coeff0 = 0
            for bid,i,j,k,coeff in lines:
                coeff0 += coeff
                t = grid_data[bid][nd][i,j,k]*coeff
                if t0 is None:
                    t0 = t 
                else:
                    t0 += t
            t0 /= coeff0
            for bid,i,j,k,coeff in lines:
                grid_data[bid][nd][i,j,k] = t0
    #周期面处理
    if periodic_pattern:
        per_pattern,per_points,vectors = periodic_pattern
        if not scalars:
            vector_set = set([j for i in vectors for j in i])
            for i in range(len(qdata[0])):
                if i not in vector_set:
                    scalars.append(i)

        for bid1, r1, theta1, bid2, r2, theta2, is_trans in per_pattern:
            i1,j1,k1 = r1 
            i2,j2,k2 = r2
            data1 = grid_data[bid1]
            data2 = grid_data[bid2]
            for nd in scalars:
                t = (data1[nd][i1,j1,k1] + data2[nd][i2,j2,k2])/2 
                data1[nd][i1,j1,k1] = t
                data2[nd][i2,j2,k2] = t
            for vdx,vdy in vectors:
                tx = data1[vdx][i1,j1,k1]*math.cos(theta1) + data1[vdy][i1,j1,k1]*math.sin(theta1)
                ty = data1[vdx][i1,j1,k1]*(-math.sin(theta1)) + data1[vdy][i1,j1,k1]*math.cos(theta1)
                tx = (data2[vdx][i2,j2,k2] + tx)/2 
                ty = (data2[vdy][i2,j2,k2] + ty)/2

                data2[vdx][i2,j2,k2] = tx
                data2[vdy][i2,j2,k2] = ty
                data1[vdx][i1,j1,k1] = tx*math.cos(theta2) + ty*math.sin(theta2)
                data1[vdy][i1,j1,k1] = tx*(-math.sin(theta2)) + ty*math.cos(theta2)
        #周期面上存在非周期块上的点时 用周期面上的点赋值
        for ps in per_points:
            bidp,ip,jp,kp,_ = ps[0]
            for nd in range(len(qdata[0])):
                ds = grid_data[bidp][nd][ip,jp,kp]
                for bid,i,j,k,_ in ps[1]:
                    grid_data[bid][nd][i,j,k] = ds

    if keys is not None:
        grid_data = [{key:val for key,val in zip(keys,q)} for q in grid_data]
        [gdata.update(adata) for gdata,adata in zip(grid_data,aux_data)]
    return grid_data
def get_lines_common(one2oneData):
    def number_to_line(data):
        _,d1 = data
        axis1,axis2 = d1[-2]-1,d1[-1]-1
        axis0 = 3-axis1-axis2
        pos1 = [None,None,None]
        pos1[axis1] = d1[axis1]
        pos1[axis2] = d1[axis2][0]
        pos1[axis0] = d1[axis0][0]

        pos2 = [None,None,None]
        pos2[axis1] = d1[axis1]
        pos2[axis2] = d1[axis2][1]
        pos2[axis0] = d1[axis0][0]

        pos3 = [None,None,None]
        pos3[axis1] = d1[axis1][0]
        pos3[axis2] = d1[axis2]
        pos3[axis0] = d1[axis0][0]

        pos4 = [None,None,None]
        pos4[axis1] = d1[axis1][1]
        pos4[axis2] = d1[axis2]
        pos4[axis0] = d1[axis0][0]
        return (pos1,axis1),(pos2,axis1),(pos3,axis2),(pos4,axis2)
    def parse_line(odata,tdata):
        if tdata[0]<odata[0]:
            tdata,odata = odata,tdata
        blockid1 = odata[0]
        blockid2 = tdata[0]

        olines = number_to_line(odata)
        tlines = number_to_line(tdata)

        return [((blockid1,l1),(blockid2,l2)) for l1,l2 in zip(olines,tlines)]

    def adjust_lines_order(lines):
        (bid1,(ijk1,axis1)),(bid2,(ijk2,axis2)) = lines
        if ijk1[axis1][0] > ijk1[axis1][1]:
            ijk1[axis1] = ijk1[axis1][1],ijk1[axis1][0]
            ijk2[axis2] = ijk2[axis2][1],ijk2[axis2][0]
        return (bid1,(tuple(ijk1),axis1)),(bid2,(tuple(ijk2),axis2))
    lineinfo2 = [adjust_lines_order(j) for i in one2oneData for j in parse_line(i[:2],i[2:])]
    lineinfo2.sort(key=lambda x:min(x[0][0],x[1][0]))
    lines_list = []
    for line1,line2 in lineinfo2:
        for lset in lines_list:
            if (line1 in lset) or (line2 in lset):
                lset.add(line1)
                lset.add(line2)
                break 
        else:
            lines_list.append(set((line1,line2)))
    return [tuple(i) for i in lines_list]
def get_line_pattern(lines_commom,shapes):
    def convert_to_slice(line,shape):
        bid,(ijk,axis) = line 
        bid -= 1
        t1,t2 = ijk[axis]
        axis1 = (axis+1)%3 
        axis2 = (axis+2)%3
        if (ijk[axis1] == 1 or ijk[axis1] == shape[axis1]) and (ijk[axis2]==1 or ijk[axis2] == shape[axis2]):
            coeff = 2 
        else:
            coeff = 4
        if t2>t1:
            t = slice(t1,t2-1)
        else:
            t = slice(t1-2,t2-1,-1)
        i = t if axis==0 else ijk[0]-1
        j = t if axis==1 else ijk[1]-1
        k = t if axis==2 else ijk[2]-1
        return bid,i,j,k,coeff
    return [[convert_to_slice(j,shapes[j[0]-1]) for j in i] for i in lines_commom]
def get_point_pattern(lines_common,shapes):
    def get_coeff(idim,jdim,kdim,i,j,k):
        s = 0
        if i==0 or i==idim:
            s+=1 
        if j==0 or j==jdim:
            s+=1
        if k==0 or k==kdim:
            s+=1
        if s==3:
            coeff = 1 
        elif s==2:
            coeff = 2 
        elif s==1:
            coeff = 4
        else:
            raise Exception('error')
        return coeff
    def split_2_points(line,shape):
        bid,(ijk,axis) = line 
        bid -= 1
        t1,t2 = ijk[axis]
        t1-=1
        t2-=1
        i1 = t1 if axis==0 else ijk[0]-1
        j1 = t1 if axis==1 else ijk[1]-1
        k1 = t1 if axis==2 else ijk[2]-1

        i2 = t2 if axis==0 else ijk[0]-1
        j2 = t2 if axis==1 else ijk[1]-1
        k2 = t2 if axis==2 else ijk[2]-1
        idim,jdim,kdim = shape 
        idim,jdim,kdim = idim-1,jdim-1,kdim-1
        coeff1 = get_coeff(idim,jdim,kdim,i1,j1,k1)
        coeff2 = get_coeff(idim,jdim,kdim,i2,j2,k2)
        return (bid,i1,j1,k1,coeff1),(bid,i2,j2,k2,coeff2)
    def is_union(point_list,ps):
        is_in = False
        for p in ps:
            for pset in point_list:
                if p in pset:
                    pset.update(ps)
                    is_in = True
                    break 
            if is_in:
                break 
        if not is_in:
            point_list.append(set(ps))
    def update_point_list(points):
        point_list = [set(points[0]) ]
        for ps in points[1:]:
            ps_in_list = False
            for p in ps:
                for i in point_list:
                    if p in i:
                        i.update(ps)
                        ps_in_list = True
                        break 
            if ps_in_list is False:
                point_list.append(set(ps))
        return point_list
    points = [kk for i in lines_common for kk in zip(*[split_2_points(j,shapes[j[0]-1]) for j in i])]
    point_result = update_point_list(points)
    return point_result
def get_periodic_pattern(periodicData,shapes):

    def convert_to_per(f1,f2,shapes):
        #bid1, r1, theta1, bid2, r2, theta2, is_trans
        bid1,faceid1,theta1 = f1[0],f1[1],f1[-1]
        bid2,faceid2,theta2 = f2[0],f2[1],f2[-1]
        r1 = [slice(None,None)]*3 
        r1[faceid1//2] =0 if faceid1 % 2 == 0 else -1 
        r2 = [slice(None,None)]*3 
        r2[faceid2//2] =0 if faceid2 % 2 == 0 else -1 
        theta1 = theta1/180*math.pi 
        theta2 = theta2/180*math.pi
        return bid1-1,r1,theta1,bid2-1,r2,theta2,False
    to_faces = [i for i in periodicData if i[-1]>0]
    from_faces = [i for i in periodicData if i[-1]<0]
    periodic_faces = []
    for i in from_faces:
        for j in to_faces:
            if i[6]==j[0]:
                periodic_faces.append(convert_to_per(i,j,shapes))
                break 
    return periodic_faces
def get_patterns(input_file,vectors=None):
    def get_face_of_line(bid,i,j,k):
        if not isinstance(i,int):
            if j==0:
                if k==0:
                    faces = 2,4
                else:
                    faces = 2,5
            else:
                if k==0:
                    faces = 3,4
                else:
                    faces = 3,5
        elif not isinstance(j,int):
            if i==0:
                if k==0:
                    faces=0,4
                else:
                    faces=0,5
            else:
                if k==0:
                    faces=1,4
                else:
                    faces=1,5
        else:
            if i==0:
                if j==0:
                    faces = 0,2
                else:
                    faces = 0,3
            else:
                if j==0:
                    faces = 1,2 
                else:
                    faces = 1,3
        return [(bid,i) for i in faces]
    def in_periodic_face(lines,per_faces,per_blocks):
        if len(lines)<=2:
            return None
        result = []
        first_line = None
        for lid,(bid,i,j,k,coeff) in enumerate(lines):
            if coeff==4:
                return None
            if (bid in per_blocks) or (first_line is not None):
                for pface in get_face_of_line(bid,i,j,k):
                    if pface in per_faces:
                        if first_line is None:
                            # input('hhhhhhhh')
                            result.extend(lines[0:lid])
                            first_line = bid,i,j,k,None
                            # print(result)
                        break
                else:
                    if first_line is not None:
                        # input('99999999999')
                        result.append((bid,i,j,k,None))
                        # print(first_line,result)

        if first_line:
            return expand_line(first_line),[expand_line(i) for i in result]
        else:
            return None
    def expand_slice(sl):
        start = sl.start 
        end = sl.stop
        step = sl.step 
        if step is None:
            step = 1 
        return slice(start-step,end+step,step)
    def expand_line(line):
        #将线扩展到周边的棱和顶点处 
        bid,i,j,k,coeff = line 
        if not isinstance(i,int):
            i = expand_slice(i)
        elif not isinstance(j,int):
            j = expand_slice(j)
        else:
            k = expand_slice(k)
        return bid,i,j,k,coeff
    one2oneData,periodicData,shapes = get_one2one_data(input_file)

    face_pattern = get_face_pattern(one2oneData)
    lines_common = get_lines_common(one2oneData)
    line_pattern = get_line_pattern(lines_common,shapes)
    point_pattern = get_point_pattern(lines_common,shapes)
    if periodicData:
        periodic_pattern = get_periodic_pattern(periodicData,shapes)
        vectors = vectors if vectors else [] 
        per_faces = set([(i[0]-1,i[1]) for i in periodicData])
        per_blocks = set([i[0]-1 for i in periodicData])
        lines_per = [in_periodic_face(i,per_faces,per_blocks) for i in line_pattern]
        lines_per = [i for i in lines_per if i]
        periodic_pattern = periodic_pattern,lines_per,vectors
    else:
        periodic_pattern = None

    return face_pattern,line_pattern,point_pattern,periodic_pattern
def auto_run(json_file = 's.json'):
    jdata = json.loads(open(json_file).read())
    patterns = get_patterns(jdata['inpfile'],vectors=jdata.get('vectors',[]))
    is_double = jdata.get('double',False)
    datafile = jdata['datafile']
    if isinstance(datafile,str):
        datafile = [datafile,]
    outfile = jdata.get('outfile',None)
    if outfile is None:
        outfile = [str(Path(i).with_name('grid_'+Path(i).name)) for i in datafile]
    for i,j  in zip(datafile,outfile):
        print(i)
        qdata = tecfile.read(i)
        t = interpolate_center2grid(qdata,patterns,is_double=is_double)
        tecfile.write(j,t)
if __name__=='__main__':
    import time 
    s = time.time()
    auto_run("/home/yxs/F/learning/CFL3D/CFL3D/Stage-B_SA_pb99240-6.7.2/s.json")
    e = time.time()
    print('time: {:.4}s'.format(e-s))